@@ -3,34 +3,64 @@
  *
  *	server checks and output routines
  *
- *	Copyright (c) 2010-2011, PostgreSQL Global Development Group
+ *	Copyright (c) 2010-2012, PostgreSQL Global Development Group
  *	contrib/pg_upgrade/check.c
  */
 
+#include "postgres.h"
+
 #include "pg_upgrade.h"
 
 
 static void set_locale_and_encoding(ClusterInfo *cluster);
 static void check_new_cluster_is_empty(void);
-static void check_old_cluster_has_new_cluster_dbs(void);
 static void check_locale_and_encoding(ControlData *oldctrl,
 						  ControlData *newctrl);
 static void check_is_super_user(ClusterInfo *cluster);
 static void check_for_prepared_transactions(ClusterInfo *cluster);
 static void check_for_isn_and_int8_passing_mismatch(ClusterInfo *cluster);
 static void check_for_reg_data_type_usage(ClusterInfo *cluster);
+static void check_for_invalid_indexes(ClusterInfo *cluster);
 static void get_bin_version(ClusterInfo *cluster);
+static char *get_canonical_locale_name(int category, const char *locale);
 
 
+/*
+ * fix_path_separator
+ * For non-Windows, just return the argument.
+ * For Windows convert any forward slash to a backslash
+ * such as is suitable for arguments to builtin commands 
+ * like RMDIR and DEL.
+ */
+static char *
+fix_path_separator(char *path)
+{
+#ifdef WIN32
+
+	char *result;
+	char *c;
+
+	result = pg_strdup(path);
+
+	for (c = result; *c != '\0'; c++)
+		if (*c == '/')
+			*c = '\\';
+
+	return result;
+
+#else
+
+	return path;
+
+#endif
+}
+
 void
 output_check_banner(bool *live_check)
 {
 	if (user_opts.check && is_server_running(old_cluster.pgdata))
 	{
 		*live_check = true;
-		if (old_cluster.port == new_cluster.port)
-			pg_log(PG_FATAL, "When checking a live server, "
-				   "the old and new port numbers must be different.\n");
 		pg_log(PG_REPORT, "Performing Consistency Checks on Old Live Server\n");
 		pg_log(PG_REPORT, "------------------------------------------------\n");
 	}
@@ -43,8 +73,7 @@ output_check_banner(bool *live_check)
 
 
 void
-check_old_cluster(bool live_check,
-				  char **sequence_script_file_name)
+check_old_cluster(bool live_check, char **sequence_script_file_name)
 {
 	/* -- OLD -- */
 
@@ -69,6 +98,7 @@ check_old_cluster(bool live_check,
 	check_is_super_user(&old_cluster);
 	check_for_prepared_transactions(&old_cluster);
 	check_for_reg_data_type_usage(&old_cluster);
+	check_for_invalid_indexes(&old_cluster);
 	check_for_isn_and_int8_passing_mismatch(&old_cluster);
 
 	/* old = PG 8.3 checks? */
@@ -124,8 +154,6 @@ check_new_cluster(void)
 
 	check_new_cluster_is_empty();
 
-	check_old_cluster_has_new_cluster_dbs();
-
 	check_loadable_libraries();
 
 	if (user_opts.transfer_mode == TRANSFER_MODE_LINK)
@@ -165,10 +193,8 @@ report_clusters_compatible(void)
 	}
 
 	pg_log(PG_REPORT, "\n"
-		   "| If pg_upgrade fails after this point, you must\n"
-		   "| re-initdb the new cluster before continuing.\n"
-		   "| You will also need to remove the \".old\" suffix\n"
-		   "| from %s/global/pg_control.old.\n", old_cluster.pgdata);
+		   "If pg_upgrade fails after this point, you must re-initdb the\n"
+		   "new cluster before continuing.\n");
 }
 
 
@@ -184,12 +210,10 @@ issue_warnings(char *sequence_script_file_name)
 		if (sequence_script_file_name)
 		{
 			prep_status("Adjusting sequences");
-			exec_prog(true,
-					  SYSTEMQUOTE "\"%s/psql\" --set ON_ERROR_STOP=on "
-					  "--no-psqlrc --port %d --username \"%s\" "
-					  "-f \"%s\" --dbname template1 >> \"%s\"" SYSTEMQUOTE,
-					  new_cluster.bindir, new_cluster.port, os_info.user,
-					  sequence_script_file_name, log_opts.filename);
+			exec_prog(UTILITY_LOG_FILE, NULL, true,
+					  "\"%s/psql\" " EXEC_PSQL_ARGS " %s -f \"%s\"",
+					  new_cluster.bindir, cluster_conn_opts(&new_cluster),
+					  sequence_script_file_name);
 			unlink(sequence_script_file_name);
 			check_ok();
 		}
@@ -211,26 +235,24 @@ issue_warnings(char *sequence_script_file_name)
 
 
 void
-output_completion_banner(char *deletion_script_file_name)
+output_completion_banner(char *analyze_script_file_name,
+						 char *deletion_script_file_name)
 {
 	/* Did we copy the free space files? */
 	if (GET_MAJOR_VERSION(old_cluster.major_version) >= 804)
 		pg_log(PG_REPORT,
-			   "| Optimizer statistics are not transferred by pg_upgrade\n"
-			   "| so consider running:\n"
-			   "| \tvacuumdb --all --analyze-only\n"
-			   "| on the newly-upgraded cluster.\n\n");
+			   "Optimizer statistics are not transferred by pg_upgrade so,\n"
+			   "once you start the new server, consider running:\n"
+			   "    %s\n\n", analyze_script_file_name);
 	else
 		pg_log(PG_REPORT,
-			   "| Optimizer statistics and free space information\n"
-			   "| are not transferred by pg_upgrade so consider\n"
-			   "| running:\n"
-			   "| \tvacuumdb --all --analyze\n"
-			   "| on the newly-upgraded cluster.\n\n");
+			   "Optimizer statistics and free space information are not transferred\n"
+		"by pg_upgrade so, once you start the new server, consider running:\n"
+			   "    %s\n\n", analyze_script_file_name);
 
 	pg_log(PG_REPORT,
-		   "| Running this script will delete the old cluster's data files:\n"
-		   "| \t%s\n",
+		   "Running this script will delete the old cluster's data files:\n"
+		   "    %s\n",
 		   deletion_script_file_name);
 }
 
@@ -259,7 +281,8 @@ check_cluster_versions(void)
 
 	/*
 	 * We can't allow downgrading because we use the target pg_dumpall, and
-	 * pg_dumpall cannot operate on new database versions, only older versions.
+	 * pg_dumpall cannot operate on new database versions, only older
+	 * versions.
 	 */
 	if (old_cluster.major_version > new_cluster.major_version)
 		pg_log(PG_FATAL, "This utility cannot be used to downgrade to older major PostgreSQL versions.\n");
@@ -295,6 +318,16 @@ check_cluster_compatibility(bool live_check)
 		new_cluster.controldata.cat_ver < TABLE_SPACE_SUBDIRS_CAT_VER)
 		pg_log(PG_FATAL, "This utility can only upgrade to PostgreSQL version 9.0 after 2010-01-11\n"
 			   "because of backend API changes made during development.\n");
+
+	/* We read the real port number for PG >= 9.1 */
+	if (live_check && GET_MAJOR_VERSION(old_cluster.major_version) < 901 &&
+		old_cluster.port == DEF_PGUPORT)
+			pg_log(PG_FATAL, "When checking a pre-PG 9.1 live old server, "
+				   "you must specify the old server's port number.\n");
+
+	if (live_check && old_cluster.port == new_cluster.port)
+		pg_log(PG_FATAL, "When checking a live server, "
+			   "the old and new port numbers must be different.\n");
 }
 
 
@@ -329,8 +362,23 @@ set_locale_and_encoding(ClusterInfo *cluster)
 		i_datcollate = PQfnumber(res, "datcollate");
 		i_datctype = PQfnumber(res, "datctype");
 
-		ctrl->lc_collate = pg_strdup(PQgetvalue(res, 0, i_datcollate));
-		ctrl->lc_ctype = pg_strdup(PQgetvalue(res, 0, i_datctype));
+		if (GET_MAJOR_VERSION(cluster->major_version) < 902)
+		{
+			/*
+			 *	Pre-9.2 did not canonicalize the supplied locale names
+			 *	to match what the system returns, while 9.2+ does, so
+			 *	convert pre-9.2 to match.
+			 */
+			ctrl->lc_collate = get_canonical_locale_name(LC_COLLATE,
+							   pg_strdup(PQgetvalue(res, 0, i_datcollate)));
+			ctrl->lc_ctype = get_canonical_locale_name(LC_CTYPE,
+							   pg_strdup(PQgetvalue(res, 0, i_datctype)));
+ 		}
+		else
+		{
+	 		ctrl->lc_collate = pg_strdup(PQgetvalue(res, 0, i_datcollate));
+			ctrl->lc_ctype = pg_strdup(PQgetvalue(res, 0, i_datctype));
+		}
 
 		PQclear(res);
 	}
@@ -360,16 +408,23 @@ static void
 check_locale_and_encoding(ControlData *oldctrl,
 						  ControlData *newctrl)
 {
-	/* These are often defined with inconsistent case, so use pg_strcasecmp(). */
+	/*
+	 *	These are often defined with inconsistent case, so use pg_strcasecmp().
+	 *	They also often use inconsistent hyphenation, which we cannot fix, e.g.
+	 *	UTF-8 vs. UTF8, so at least we display the mismatching values.
+	 */
 	if (pg_strcasecmp(oldctrl->lc_collate, newctrl->lc_collate) != 0)
 		pg_log(PG_FATAL,
-			   "old and new cluster lc_collate values do not match\n");
+			   "lc_collate cluster values do not match:  old \"%s\", new \"%s\"\n",
+			   oldctrl->lc_collate, newctrl->lc_collate);
 	if (pg_strcasecmp(oldctrl->lc_ctype, newctrl->lc_ctype) != 0)
 		pg_log(PG_FATAL,
-			   "old and new cluster lc_ctype values do not match\n");
+			   "lc_ctype cluster values do not match:  old \"%s\", new \"%s\"\n",
+			   oldctrl->lc_ctype, newctrl->lc_ctype);
 	if (pg_strcasecmp(oldctrl->encoding, newctrl->encoding) != 0)
 		pg_log(PG_FATAL,
-			   "old and new cluster encoding values do not match\n");
+			   "encoding cluster values do not match:  old \"%s\", new \"%s\"\n",
+			   oldctrl->encoding, newctrl->encoding);
 }
 
 
@@ -397,30 +452,129 @@ check_new_cluster_is_empty(void)
 
 
 /*
- *	If someone removes the 'postgres' database from the old cluster and
- *	the new cluster has a 'postgres' database, the number of databases
- *	will not match.  We actually could upgrade such a setup, but it would
- *	violate the 1-to-1 mapping of database counts, so we throw an error
- *	instead.  We would detect this as a database count mismatch during
- *	upgrade, but we want to detect it during the check phase and report
- *	the database name.
+ * create_script_for_cluster_analyze()
+ *
+ *	This incrementally generates better optimizer statistics
  */
-static void
-check_old_cluster_has_new_cluster_dbs(void)
+void
+create_script_for_cluster_analyze(char **analyze_script_file_name)
 {
-	int			old_dbnum,
-				new_dbnum;
+	FILE	   *script = NULL;
 
-	for (new_dbnum = 0; new_dbnum < new_cluster.dbarr.ndbs; new_dbnum++)
-	{
-		for (old_dbnum = 0; old_dbnum < old_cluster.dbarr.ndbs; old_dbnum++)
-			if (strcmp(old_cluster.dbarr.dbs[old_dbnum].db_name,
-					   new_cluster.dbarr.dbs[new_dbnum].db_name) == 0)
-				break;
-		if (old_dbnum == old_cluster.dbarr.ndbs)
-			pg_log(PG_FATAL, "New cluster database \"%s\" does not exist in the old cluster\n",
-				   new_cluster.dbarr.dbs[new_dbnum].db_name);
-	}
+	*analyze_script_file_name = pg_malloc(MAXPGPATH);
+
+	prep_status("Creating script to analyze new cluster");
+
+	snprintf(*analyze_script_file_name, MAXPGPATH, "analyze_new_cluster.%s",
+			 SCRIPT_EXT);
+
+	if ((script = fopen_priv(*analyze_script_file_name, "w")) == NULL)
+		pg_log(PG_FATAL, "Could not open file \"%s\": %s\n",
+			   *analyze_script_file_name, getErrorText(errno));
+
+#ifndef WIN32
+	/* add shebang header */
+	fprintf(script, "#!/bin/sh\n\n");
+#else
+	/* suppress command echoing */
+	fprintf(script, "@echo off\n");
+#endif
+
+	fprintf(script, "echo %sThis script will generate minimal optimizer statistics rapidly%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %sso your system is usable, and then gather statistics twice more%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %swith increasing accuracy.  When it is done, your system will%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %shave the default level of optimizer statistics.%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo%s\n\n", ECHO_BLANK);
+
+	fprintf(script, "echo %sIf you have used ALTER TABLE to modify the statistics target for%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %sany tables, you might want to remove them and restore them after%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %srunning this script because they will delay fast statistics generation.%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo%s\n\n", ECHO_BLANK);
+
+	fprintf(script, "echo %sIf you would like default statistics as quickly as possible, cancel%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %sthis script and run:%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %s    \"%s/vacuumdb\" --all %s%s\n", ECHO_QUOTE, new_cluster.bindir,
+	/* Did we copy the free space files? */
+			(GET_MAJOR_VERSION(old_cluster.major_version) >= 804) ?
+			"--analyze-only" : "--analyze", ECHO_QUOTE);
+	fprintf(script, "echo%s\n\n", ECHO_BLANK);
+
+#ifndef WIN32
+	fprintf(script, "sleep 2\n");
+	fprintf(script, "PGOPTIONS='-c default_statistics_target=1 -c vacuum_cost_delay=0'\n");
+	/* only need to export once */
+	fprintf(script, "export PGOPTIONS\n");
+#else
+	fprintf(script, "REM simulate sleep 2\n");
+	fprintf(script, "PING 1.1.1.1 -n 1 -w 2000 > nul\n");
+	fprintf(script, "SET PGOPTIONS=-c default_statistics_target=1 -c vacuum_cost_delay=0\n");
+#endif
+
+	fprintf(script, "echo %sGenerating minimal optimizer statistics (1 target)%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %s--------------------------------------------------%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "\"%s/vacuumdb\" --all --analyze-only\n", new_cluster.bindir);
+	fprintf(script, "echo%s\n", ECHO_BLANK);
+	fprintf(script, "echo %sThe server is now available with minimal optimizer statistics.%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %sQuery performance will be optimal once this script completes.%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo%s\n\n", ECHO_BLANK);
+
+#ifndef WIN32
+	fprintf(script, "sleep 2\n");
+	fprintf(script, "PGOPTIONS='-c default_statistics_target=10'\n");
+#else
+	fprintf(script, "REM simulate sleep\n");
+	fprintf(script, "PING 1.1.1.1 -n 1 -w 2000 > nul\n");
+	fprintf(script, "SET PGOPTIONS=-c default_statistics_target=10\n");
+#endif
+
+	fprintf(script, "echo %sGenerating medium optimizer statistics (10 targets)%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %s---------------------------------------------------%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "\"%s/vacuumdb\" --all --analyze-only\n", new_cluster.bindir);
+	fprintf(script, "echo%s\n\n", ECHO_BLANK);
+
+#ifndef WIN32
+	fprintf(script, "unset PGOPTIONS\n");
+#else
+	fprintf(script, "SET PGOPTIONS\n");
+#endif
+
+	fprintf(script, "echo %sGenerating default (full) optimizer statistics (100 targets?)%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "echo %s-------------------------------------------------------------%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+	fprintf(script, "\"%s/vacuumdb\" --all %s\n", new_cluster.bindir,
+	/* Did we copy the free space files? */
+			(GET_MAJOR_VERSION(old_cluster.major_version) >= 804) ?
+			"--analyze-only" : "--analyze");
+
+	fprintf(script, "echo%s\n\n", ECHO_BLANK);
+	fprintf(script, "echo %sDone%s\n",
+			ECHO_QUOTE, ECHO_QUOTE);
+
+	fclose(script);
+
+#ifndef WIN32
+	if (chmod(*analyze_script_file_name, S_IRWXU) != 0)
+		pg_log(PG_FATAL, "Could not add execute permission to file \"%s\": %s\n",
+			   *analyze_script_file_name, getErrorText(errno));
+#endif
+
+	check_ok();
 }
 
 
@@ -430,8 +584,7 @@ check_old_cluster_has_new_cluster_dbs(void)
  *	This is particularly useful for tablespace deletion.
  */
 void
-create_script_for_old_cluster_deletion(
-									   char **deletion_script_file_name)
+create_script_for_old_cluster_deletion(char **deletion_script_file_name)
 {
 	FILE	   *script = NULL;
 	int			tblnum;
@@ -440,12 +593,12 @@ create_script_for_old_cluster_deletion(
 
 	prep_status("Creating script to delete old cluster");
 
-	snprintf(*deletion_script_file_name, MAXPGPATH, "%s/delete_old_cluster.%s",
-			 os_info.cwd, SCRIPT_EXT);
+	snprintf(*deletion_script_file_name, MAXPGPATH, "delete_old_cluster.%s",
+			 SCRIPT_EXT);
 
-	if ((script = fopen(*deletion_script_file_name, "w")) == NULL)
-		pg_log(PG_FATAL, "Could not create necessary file:  %s\n",
-			   *deletion_script_file_name);
+	if ((script = fopen_priv(*deletion_script_file_name, "w")) == NULL)
+		pg_log(PG_FATAL, "Could not open file \"%s\": %s\n",
+			   *deletion_script_file_name, getErrorText(errno));
 
 #ifndef WIN32
 	/* add shebang header */
@@ -453,7 +606,7 @@ create_script_for_old_cluster_deletion(
 #endif
 
 	/* delete old cluster's default tablespace */
-	fprintf(script, RMDIR_CMD " %s\n", old_cluster.pgdata);
+	fprintf(script, RMDIR_CMD " %s\n", fix_path_separator(old_cluster.pgdata));
 
 	/* delete old cluster's alternate tablespaces */
 	for (tblnum = 0; tblnum < os_info.num_tablespaces; tblnum++)
@@ -470,14 +623,17 @@ create_script_for_old_cluster_deletion(
 			fprintf(script, "\n");
 			/* remove PG_VERSION? */
 			if (GET_MAJOR_VERSION(old_cluster.major_version) <= 804)
-				fprintf(script, RM_CMD " %s%s/PG_VERSION\n",
-				 os_info.tablespaces[tblnum], old_cluster.tablespace_suffix);
+				fprintf(script, RM_CMD " %s%s%cPG_VERSION\n",
+						fix_path_separator(os_info.tablespaces[tblnum]), 
+						fix_path_separator(old_cluster.tablespace_suffix),
+						PATH_SEPARATOR);
 
-			for (dbnum = 0; dbnum < new_cluster.dbarr.ndbs; dbnum++)
+			for (dbnum = 0; dbnum < old_cluster.dbarr.ndbs; dbnum++)
 			{
-				fprintf(script, RMDIR_CMD " %s%s/%d\n",
-				  os_info.tablespaces[tblnum], old_cluster.tablespace_suffix,
-						old_cluster.dbarr.dbs[dbnum].db_oid);
+				fprintf(script, RMDIR_CMD " %s%s%c%d\n",
+						fix_path_separator(os_info.tablespaces[tblnum]),
+						fix_path_separator(old_cluster.tablespace_suffix),
+						PATH_SEPARATOR, old_cluster.dbarr.dbs[dbnum].db_oid);
 			}
 		}
 		else
@@ -487,15 +643,16 @@ create_script_for_old_cluster_deletion(
 			 * or a version-specific subdirectory.
 			 */
 			fprintf(script, RMDIR_CMD " %s%s\n",
-				 os_info.tablespaces[tblnum], old_cluster.tablespace_suffix);
+					fix_path_separator(os_info.tablespaces[tblnum]), 
+					fix_path_separator(old_cluster.tablespace_suffix));
 	}
 
 	fclose(script);
 
 #ifndef WIN32
 	if (chmod(*deletion_script_file_name, S_IRWXU) != 0)
-		pg_log(PG_FATAL, "Could not add execute permission to file:  %s\n",
-			   *deletion_script_file_name);
+		pg_log(PG_FATAL, "Could not add execute permission to file \"%s\": %s\n",
+			   *deletion_script_file_name, getErrorText(errno));
 #endif
 
 	check_ok();
@@ -601,8 +758,8 @@ check_for_isn_and_int8_passing_mismatch(ClusterInfo *cluster)
 		return;
 	}
 
-	snprintf(output_path, sizeof(output_path), "%s/contrib_isn_and_int8_pass_by_value.txt",
-			 os_info.cwd);
+	snprintf(output_path, sizeof(output_path),
+			 "contrib_isn_and_int8_pass_by_value.txt");
 
 	for (dbnum = 0; dbnum < cluster->dbarr.ndbs; dbnum++)
 	{
@@ -629,11 +786,12 @@ check_for_isn_and_int8_passing_mismatch(ClusterInfo *cluster)
 		for (rowno = 0; rowno < ntups; rowno++)
 		{
 			found = true;
-			if (script == NULL && (script = fopen(output_path, "w")) == NULL)
-				pg_log(PG_FATAL, "Could not create necessary file:  %s\n", output_path);
+			if (script == NULL && (script = fopen_priv(output_path, "w")) == NULL)
+				pg_log(PG_FATAL, "Could not open file \"%s\": %s\n",
+					   output_path, getErrorText(errno));
 			if (!db_used)
 			{
-				fprintf(script, "Database:  %s\n", active_db->db_name);
+				fprintf(script, "Database: %s\n", active_db->db_name);
 				db_used = true;
 			}
 			fprintf(script, "  %s.%s\n",
@@ -653,15 +811,13 @@ check_for_isn_and_int8_passing_mismatch(ClusterInfo *cluster)
 	{
 		pg_log(PG_REPORT, "fatal\n");
 		pg_log(PG_FATAL,
-			   "| Your installation contains \"contrib/isn\" functions\n"
-			   "| which rely on the bigint data type.  Your old and\n"
-			   "| new clusters pass bigint values differently so this\n"
-			   "| cluster cannot currently be upgraded.  You can\n"
-			   "| manually upgrade data that use \"contrib/isn\"\n"
-			   "| facilities and remove \"contrib/isn\" from the\n"
-			   "| old cluster and restart the upgrade.  A list\n"
-			   "| of the problem functions is in the file:\n"
-			   "| \t%s\n\n", output_path);
+			   "Your installation contains \"contrib/isn\" functions which rely on the\n"
+		  "bigint data type.  Your old and new clusters pass bigint values\n"
+		"differently so this cluster cannot currently be upgraded.  You can\n"
+			   "manually upgrade databases that use \"contrib/isn\" facilities and remove\n"
+			   "\"contrib/isn\" from the old cluster and restart the upgrade.  A list of\n"
+			   "the problem functions is in the file:\n"
+			   "    %s\n\n", output_path);
 	}
 	else
 		check_ok();
@@ -671,11 +827,11 @@ check_for_isn_and_int8_passing_mismatch(ClusterInfo *cluster)
 /*
  * check_for_reg_data_type_usage()
  *	pg_upgrade only preserves these system values:
- *		pg_class.relfilenode
+ *		pg_class.oid
  *		pg_type.oid
  *		pg_enum.oid
  *
- *	Most of the reg* data types reference system catalog info that is
+ *	Many of the reg* data types reference system catalog info that is
  *	not preserved, and hence these data types cannot be used in user
  *	tables upgraded by pg_upgrade.
  */
@@ -687,10 +843,9 @@ check_for_reg_data_type_usage(ClusterInfo *cluster)
 	bool		found = false;
 	char		output_path[MAXPGPATH];
 
-	prep_status("Checking for reg* system oid user data types");
+	prep_status("Checking for reg* system OID user data types");
 
-	snprintf(output_path, sizeof(output_path), "%s/tables_using_reg.txt",
-			 os_info.cwd);
+	snprintf(output_path, sizeof(output_path), "tables_using_reg.txt");
 
 	for (dbnum = 0; dbnum < cluster->dbarr.ndbs; dbnum++)
 	{
@@ -704,6 +859,11 @@ check_for_reg_data_type_usage(ClusterInfo *cluster)
 		DbInfo	   *active_db = &cluster->dbarr.dbs[dbnum];
 		PGconn	   *conn = connectToServer(cluster, active_db->db_name);
 
+		/*
+		 * While several relkinds don't store any data, e.g. views, they can
+		 * be used to define data types of other columns, so we check all
+		 * relkinds.
+		 */
 		res = executeQueryOrDie(conn,
 								"SELECT n.nspname, c.relname, a.attname "
 								"FROM	pg_catalog.pg_class c, "
@@ -716,7 +876,7 @@ check_for_reg_data_type_usage(ClusterInfo *cluster)
 								"			'pg_catalog.regprocedure'::pg_catalog.regtype, "
 		  "			'pg_catalog.regoper'::pg_catalog.regtype, "
 								"			'pg_catalog.regoperator'::pg_catalog.regtype, "
-/*	allow	 "			'pg_catalog.regclass'::pg_catalog.regtype, "*/
+		/* regclass.oid is preserved, so 'regclass' is OK */
 		/* regtype.oid is preserved, so 'regtype' is OK */
 		"			'pg_catalog.regconfig'::pg_catalog.regtype, "
 								"			'pg_catalog.regdictionary'::pg_catalog.regtype) AND "
@@ -731,11 +891,12 @@ check_for_reg_data_type_usage(ClusterInfo *cluster)
 		for (rowno = 0; rowno < ntups; rowno++)
 		{
 			found = true;
-			if (script == NULL && (script = fopen(output_path, "w")) == NULL)
-				pg_log(PG_FATAL, "Could not create necessary file:  %s\n", output_path);
+			if (script == NULL && (script = fopen_priv(output_path, "w")) == NULL)
+				pg_log(PG_FATAL, "Could not open file \"%s\": %s\n",
+					   output_path, getErrorText(errno));
 			if (!db_used)
 			{
-				fprintf(script, "Database:  %s\n", active_db->db_name);
+				fprintf(script, "Database: %s\n", active_db->db_name);
 				db_used = true;
 			}
 			fprintf(script, "  %s.%s.%s\n",
@@ -756,13 +917,101 @@ check_for_reg_data_type_usage(ClusterInfo *cluster)
 	{
 		pg_log(PG_REPORT, "fatal\n");
 		pg_log(PG_FATAL,
-			   "| Your installation contains one of the reg* data types in\n"
-			   "| user tables.  These data types reference system oids that\n"
-			   "| are not preserved by pg_upgrade, so this cluster cannot\n"
-			   "| currently be upgraded.  You can remove the problem tables\n"
-			   "| and restart the upgrade.  A list of the problem columns\n"
-			   "| is in the file:\n"
-			   "| \t%s\n\n", output_path);
+			   "Your installation contains one of the reg* data types in user tables.\n"
+		 "These data types reference system OIDs that are not preserved by\n"
+		"pg_upgrade, so this cluster cannot currently be upgraded.  You can\n"
+			   "remove the problem tables and restart the upgrade.  A list of the problem\n"
+			   "columns is in the file:\n"
+			   "    %s\n\n", output_path);
+	}
+	else
+		check_ok();
+}
+
+
+/*
+ * check_for_invalid_indexes()
+ *
+ *	CREATE INDEX CONCURRENTLY can create invalid indexes if the index build
+ *	fails.  These are dumped as valid indexes by pg_dump, but the
+ *	underlying files are still invalid indexes.  This checks to make sure
+ *	no invalid indexes exist, either failed index builds or concurrent
+ *	indexes in the process of being created.
+ */
+static void
+check_for_invalid_indexes(ClusterInfo *cluster)
+{
+	int			dbnum;
+	FILE	   *script = NULL;
+	bool		found = false;
+	char		output_path[MAXPGPATH];
+
+	prep_status("Checking for invalid indexes from concurrent index builds");
+
+	snprintf(output_path, sizeof(output_path), "invalid_indexes.txt");
+
+	for (dbnum = 0; dbnum < cluster->dbarr.ndbs; dbnum++)
+	{
+		PGresult   *res;
+		bool		db_used = false;
+		int			ntups;
+		int			rowno;
+		int			i_nspname,
+					i_relname;
+		DbInfo	   *active_db = &cluster->dbarr.dbs[dbnum];
+		PGconn	   *conn = connectToServer(cluster, active_db->db_name);
+
+		res = executeQueryOrDie(conn,
+								"SELECT n.nspname, c.relname "
+								"FROM	pg_catalog.pg_class c, "
+								"		pg_catalog.pg_namespace n, "
+								"		pg_catalog.pg_index i "
+								"WHERE	(i.indisvalid = false OR "
+								"		 i.indisready = false) AND "
+								"		i.indexrelid = c.oid AND "
+								"		c.relnamespace = n.oid AND "
+								/* we do not migrate these, so skip them */
+							    " 		n.nspname != 'pg_catalog' AND "
+								"		n.nspname != 'information_schema' AND "
+								/* indexes do not have toast tables */
+								"		n.nspname != 'pg_toast'");
+
+		ntups = PQntuples(res);
+		i_nspname = PQfnumber(res, "nspname");
+		i_relname = PQfnumber(res, "relname");
+		for (rowno = 0; rowno < ntups; rowno++)
+		{
+			found = true;
+			if (script == NULL && (script = fopen_priv(output_path, "w")) == NULL)
+				pg_log(PG_FATAL, "Could not open file \"%s\": %s\n",
+					   output_path, getErrorText(errno));
+			if (!db_used)
+			{
+				fprintf(script, "Database: %s\n", active_db->db_name);
+				db_used = true;
+			}
+			fprintf(script, "  %s.%s\n",
+					PQgetvalue(res, rowno, i_nspname),
+					PQgetvalue(res, rowno, i_relname));
+		}
+
+		PQclear(res);
+
+		PQfinish(conn);
+	}
+
+	if (script)
+		fclose(script);
+
+	if (found)
+	{
+		pg_log(PG_REPORT, "fatal\n");
+		pg_log(PG_FATAL,
+			   "Your installation contains invalid indexes due to failed or\n"
+		 	   "currently running CREATE INDEX CONCURRENTLY operations.  You\n"
+			   "cannot upgrade until these indexes are valid or removed.  A\n"
+			   "list of the problem indexes is in the file:\n"
+			   "    %s\n\n", output_path);
 	}
 	else
 		check_ok();
@@ -772,17 +1021,18 @@ check_for_reg_data_type_usage(ClusterInfo *cluster)
 static void
 get_bin_version(ClusterInfo *cluster)
 {
-	char		cmd[MAXPGPATH], cmd_output[MAX_STRING];
+	char		cmd[MAXPGPATH],
+				cmd_output[MAX_STRING];
 	FILE	   *output;
-	int			pre_dot, post_dot;
+	int			pre_dot,
+				post_dot;
 
 	snprintf(cmd, sizeof(cmd), "\"%s/pg_ctl\" --version", cluster->bindir);
 
-	if ((output = popen(cmd, "r")) == NULL)
-		pg_log(PG_FATAL, "Could not get pg_ctl version data: %s\n",
-			   getErrorText(errno));
-
-	fgets(cmd_output, sizeof(cmd_output), output);
+	if ((output = popen(cmd, "r")) == NULL ||
+		fgets(cmd_output, sizeof(cmd_output), output) == NULL)
+		pg_log(PG_FATAL, "Could not get pg_ctl version data using %s: %s\n",
+			   cmd, getErrorText(errno));
 
 	pclose(output);
 
@@ -796,3 +1046,39 @@ get_bin_version(ClusterInfo *cluster)
 	cluster->bin_version = (pre_dot * 100 + post_dot) * 100;
 }
 
+
+/*
+ * get_canonical_locale_name
+ *
+ * Send the locale name to the system, and hope we get back a canonical
+ * version.  This should match the backend's check_locale() function.
+ */
+static char *
+get_canonical_locale_name(int category, const char *locale)
+{
+	char	   *save;
+	char	   *res;
+
+	save = setlocale(category, NULL);
+	if (!save)
+        pg_log(PG_FATAL, "failed to get the current locale\n");
+
+	/* 'save' may be pointing at a modifiable scratch variable, so copy it. */
+	save = pg_strdup(save);
+
+	/* set the locale with setlocale, to see if it accepts it. */
+	res = setlocale(category, locale);
+
+	if (!res)
+        pg_log(PG_FATAL, "failed to get system local name for \"%s\"\n", res);
+
+	res = pg_strdup(res);
+
+	/* restore old value. */
+	if (!setlocale(category, save))
+        pg_log(PG_FATAL, "failed to restore old locale \"%s\"\n", save);
+
+	free(save);
+
+	return res;
+}
