@@ -5,11 +5,11 @@
  *	  clients and standalone backends are supported here).
  *
  *
- * Portions Copyright (c) 1996-2010, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/access/common/printtup.c,v 1.106 2010/01/02 16:57:33 momjian Exp $
+ *	  src/backend/access/common/printtup.c
  *
  *-------------------------------------------------------------------------
  */
